package profiler;

import ij.IJ;
import ij.ImageListener;
import ij.ImagePlus;
import ij.io.FileInfo;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneConstants;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import shared.GUV;
import shared.GUVReader;
import shared.IJFrame;

/**Main UI view.*/
public class Dialog extends IJFrame implements ImageListener, PropertyChangeListener {

   private final Profiler profiler;
   private final KeyAdapter keyAdapter;
   private final MouseAdapter mouseAdapter;
   private final JTextField imageField;
   private final JCheckBox displayOutlines;
   private final JTable guvTable, profileTable;
   private final GUVTableModel guvTableModel;
   private final ProfileTableModel profileTableModel;
   private final JFormattedTextField meanField, corrField, sdField;
   private final Border highlight = BorderFactory.createLineBorder(Color.MAGENTA, 1);

   public Dialog(Profiler profiler, String title, String version) {
      super(title, version);
      this.profiler = profiler;
      profiler.addListener(this);
      //Our own "quit" method is not called if the ImageJ frame is closed rather than ours
      ij.addWindowListener(new WindowAdapter() {

         @Override
         public void windowClosing(WindowEvent e) {
            GUV.savePreferences();
         }
      });
      ImagePlus.addImageListener(this);
      keyAdapter = new KeyAdapter() {

         @Override
         public void keyTyped(KeyEvent event) {
            if (event.getKeyChar() == 'a') {
               add();
            }
         }
      };
      mouseAdapter = new MouseAdapter() {

         @Override
         public void mouseClicked(MouseEvent event) {
            int pIndex = profileTable.getSelectedRow();
            if (pIndex > -1 && IJ.getToolName().equals("point")) {
               int gIndex = guvTable.getSelectedRow();
               profiler.adjustLine(gIndex, pIndex);
               profileTableModel.fireTableRowsUpdated(pIndex, pIndex);
               guvTableModel.fireTableRowsUpdated(gIndex, gIndex);
               profiler.displayProfile(gIndex, pIndex);
            }
         }
      };

      //Layout
      JPanel cp = new JPanel(new GridBagLayout());
      setContentPane(cp);
      cp.setBorder(new EmptyBorder(5, 5, 5, 5));

      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.LINE_START;
      c.insets = new Insets(2, 2, 2, 2);
      c.weightx = 0;
      c.weighty = 0;

      c.gridx = 0;
      c.gridy = 0;
      JLabel imageLabel = new JLabel("Image");
      imageLabel.setBorder(new EmptyBorder(0, 0, 0, 10));
      Font bold = imageLabel.getFont().deriveFont(Font.BOLD);
      imageLabel.setFont(bold);
      cp.add(imageLabel, c);

      c.gridx ++;
      c.gridwidth = 2;
      c.fill = GridBagConstraints.HORIZONTAL;
      imageField = new JTextField("");
      imageField.setColumns(15);
      cp.add(imageField, c);
      constraintReset(c);

      c.gridx = 1;
      c.gridy ++;
      c.gridwidth = 2;
      JPanel imageButtonPanel = new JPanel(new GridLayout(1, 4));
      JButton setImageButton = new JButton("Set");
      setImageButton.setToolTipText("Sets the current image active");
      setImageButton.addActionListener(e -> setImage(IJ.getImage()));
      imageButtonPanel.add(setImageButton);
      JButton drawButton = new JButton("Draw");
      drawButton.setToolTipText("Activates the Blow/Lasso tool");
      drawButton.addActionListener(e -> {
         if (!IJ.setTool("Lasso/Blow Tool")) {
            IJ.run("Blow/Lasso Tool");
         }
      });
      imageButtonPanel.add(drawButton);
      JButton addButton = new JButton("Add [a]");
      addButton.setToolTipText("Click to add a new ROI or type \"a\" while the mouse is over the image");
      addButton.addActionListener(e -> add());
      imageButtonPanel.add(addButton);

      JButton clearButton = new JButton("Clear");
      clearButton.setToolTipText("Deletes all results and inactivates the current image");
      clearButton.addActionListener(new ActionListener() {

         @Override
         public void actionPerformed(ActionEvent e) {
            profileTable.getSelectionModel().clearSelection();
            guvTable.getSelectionModel().clearSelection();
            if (profiler.imp != null) {
               profiler.imp.getCanvas().removeKeyListener(keyAdapter);
            }
            profiler.reset(false);
            imageField.setText("");
            guvTableModel.reset();
            profileTableModel.fireTableDataChanged();
         }
      });
      imageButtonPanel.add(clearButton);
      cp.add(imageButtonPanel, c);
      constraintReset(c);

      c.gridx = 1;
      c.gridy ++;
      c.gridwidth = 2;
      JPanel parameterPanel = new JPanel(new FlowLayout(FlowLayout.LEADING));
      parameterPanel.add(new JLabel("Profiles"));
      SpinnerNumberModel profileSpinnerModel = new SpinnerNumberModel(GUV.nProfiles,
            4, 30, 1);
      JSpinner profileSpinner = new JSpinner(profileSpinnerModel);
      profileSpinner.addChangeListener(e -> GUV.nProfiles = (int) profileSpinnerModel.getNumber());
      parameterPanel.add(profileSpinner);
      parameterPanel.add(Box.createHorizontalStrut(5));
      parameterPanel.add(new JLabel("Extend by"));
      SpinnerNumberModel extSpinnerModel = new SpinnerNumberModel(GUV.profileExt,
            3, 20, 1);
      JSpinner extSpinner = new JSpinner(extSpinnerModel);
      extSpinner.addChangeListener(e -> GUV.profileExt = (int) extSpinnerModel.getNumber());
      parameterPanel.add(extSpinner);
      parameterPanel.add(new JLabel("µm"));
      parameterPanel.add(Box.createHorizontalStrut(5));
      displayOutlines = new JCheckBox("Display outlines", true);
      displayOutlines.addItemListener(event -> {
         int change = event.getStateChange();
         if (change == ItemEvent.SELECTED) {
            profiler.displayOutlines(true);
         } else if (change == ItemEvent.DESELECTED) {
            profiler.displayOutlines(false);
         }
      });
      parameterPanel.add(displayOutlines);
      cp.add(parameterPanel, c);
      constraintReset(c);

      c.gridx = 0;
      c.gridy ++;
      cp.add(Box.createVerticalStrut(25));

      //Table titles
      c.gridx = 0;
      c.gridy ++;
      c.gridwidth = 3;
      c.fill = GridBagConstraints.HORIZONTAL;
      JPanel tableTitlePanel = new JPanel(new GridLayout(1, 2));
      JLabel vesicleLabel = new JLabel("Vesicles");
      vesicleLabel.setFont(bold);
      tableTitlePanel.add(vesicleLabel, c);
      JLabel profileLabel = new JLabel("Profiles");
      profileLabel.setFont(bold);
      tableTitlePanel.add(profileLabel);
      cp.add(tableTitlePanel, c);
      constraintReset(c);

      //Tables
      c.gridx = 0;
      c.gridy ++;
      c.gridwidth = 3;
      JPanel tablePanel = new JPanel();
      guvTableModel = profiler.guvTableModel;
      guvTable = new JTable(guvTableModel) { //paints a background GUV in a different color

         private final Border empty = BorderFactory.createEmptyBorder();

         @Override
         public Component prepareRenderer(TableCellRenderer renderer, int row, int column) {
            JComponent c = (JComponent) super.prepareRenderer(renderer, row, column);
            if (this.convertRowIndexToModel(row) == profiler.getControl()) {
               c.setBorder(highlight);
            } else {
               c.setBorder(empty);
            }
            return c;
         }
      };
      guvTable.getSelectionModel().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      guvTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {

         @Override
         public void valueChanged(ListSelectionEvent event) {
            if (!event.getValueIsAdjusting()) {
               int index = guvTable.getSelectedRow();
               if (index < 0) { //no selection
                  profileTable.clearSelection();
               } else {
                  index = guvTable.convertRowIndexToModel(index);
               }
               profileTableModel.setGUV(index);
               profiler.displayProfiles(index, displayOutlines.isSelected());
            }
         }
      });
      int rowHeight = guvTable.getRowHeight() + 4;
      int rowMargin = guvTable.getRowMargin() + 2;
      guvTable.setRowHeight(rowHeight);
      guvTable.setRowMargin(rowMargin);
      guvTable.setShowGrid(false);
      guvTable.setAutoCreateRowSorter(true);
      guvTable.setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);
      HeaderRenderer vesicleHeaderRenderer = new HeaderRenderer(guvTable, SwingConstants.CENTER);
      DoubleRenderer doubleRenderer = new DoubleRenderer();
      for (int i = 0; i < guvTableModel.getColumnCount(); i++) {
         TableColumn column = guvTable.getColumn(guvTableModel.getColumnName(i));
         column.setHeaderRenderer(vesicleHeaderRenderer);
         switch (i) {
            case 0: columnWidth(column, 30); break;
            case 1: columnWidth(column, 45); break;
            case 2:
            case 3:
               columnWidth(column, 80);
               column.setCellRenderer(doubleRenderer); break;
         }
      }
      JPopupMenu popup = new JPopupMenu();
      JMenuItem deleteSelected = new JMenuItem("Delete selected");
      deleteSelected.addActionListener(e -> {
         int index = guvTable.getSelectedRow();
         if (index > -1) {
            profiler.delete(guvTable.convertRowIndexToModel(index));
            guvTable.getSelectionModel().clearSelection();
            guvTableModel.fireTableDataChanged();
         }
      });
      popup.add(deleteSelected);
      JMenuItem deleteAll = new JMenuItem("Delete all");
      deleteAll.addActionListener(e -> {
         profiler.deleteAll();
         guvTableModel.fireTableDataChanged();
      });
      popup.add(deleteAll);
      JMenuItem setAsControl = new JMenuItem("Set as control");
      setAsControl.addActionListener(e -> {
         int previous = profiler.getControl();
         int index = guvTable.convertRowIndexToModel(guvTable.getSelectedRow());
         profiler.setAsControl(index);
         guvTableModel.fireTableDataChanged();
         if (previous >= 0) {
            Object[] options = {"Yes", "No"};
            int value = JOptionPane.showOptionDialog(this,
                  "Delete the previous control?",Profiler.TITLE, JOptionPane.YES_NO_OPTION,
                  JOptionPane.QUESTION_MESSAGE,null,options, options[0]);
            if (value == 0) {
               profiler.delete(previous);
               guvTableModel.fireTableDataChanged();
            }
         }
         guvTable.repaint();
      });
      popup.add(setAsControl);
      guvTable.setComponentPopupMenu(popup);
      Dimension tableSize = new Dimension(guvTable.getPreferredSize().width, 250);
      guvTable.setMinimumSize(tableSize);
      guvTable.setPreferredScrollableViewportSize(tableSize);
      JScrollPane guvTableScroll = new JScrollPane(guvTable);
      guvTableScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
      tablePanel.add(guvTableScroll);

      profileTableModel = profiler.profileTableModel;
      profileTable = new JTable(profileTableModel);
      profileTable.getSelectionModel().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      profileTable.setRowHeight(rowHeight);
      profileTable.setRowMargin(rowMargin);
      profileTable.setShowGrid(false);
      profileTable.setAutoCreateRowSorter(true);
      profileTable.setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);
      profileTable.getSelectionModel().addListSelectionListener(event -> {
         if (!event.getValueIsAdjusting()) {
            int profile = profileTable.getSelectedRow();
            if (profile < 0) { //no selection
               profileTable.clearSelection();
            } else {
               profile = profileTable.convertRowIndexToModel(profile);
               int index = guvTable.convertRowIndexToModel(guvTable.getSelectedRow());
               profiler.displayProfile(index, profile);
            }
         }
      });
      HeaderRenderer profileHeaderRenderer = new HeaderRenderer(profileTable, SwingConstants.CENTER);
      for (int i = 0; i < profileTableModel.getColumnCount(); i++) {
         TableColumn column = profileTable.getColumn(profileTableModel.getColumnName(i));
         column.setHeaderRenderer(profileHeaderRenderer);
         switch (i) {
            case 0: columnWidth(column, 30); break;
            case 1: columnWidth(column, 45); break;
            case 2: columnWidth(column, 80);
               column.setCellRenderer(doubleRenderer); break;
            case 3:
               columnWidth(column, 80); break;
         }
      }
      tableSize.width = profileTable.getPreferredSize().width;
      profileTable.setMinimumSize(tableSize);
      profileTable.setPreferredScrollableViewportSize(tableSize);
      JScrollPane profileTableScroll = new JScrollPane(profileTable);
      profileTableScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
      tablePanel.add(profileTableScroll);
      cp.add(tablePanel, c);
      constraintReset(c);

      //Statistics
      c.gridx = 0;
      c.gridy ++;
      c.gridwidth = 4;
      c.fill = GridBagConstraints.NONE;
      JPanel statsPanel = new JPanel(new GridLayout(2, 4, 3, 0));
      JLabel mean = new JLabel("Mean raw");
      mean.setFont(bold);
      statsPanel.add(mean);
      JLabel corr = new JLabel("Corrected");
      corr.setFont(bold);
      statsPanel.add(corr);
      JLabel sd = new JLabel(" +/-SD");
      sd.setFont(bold);
      statsPanel.add(sd);
      statsPanel.add(new JLabel("")); //just a dummy for the grid panel
      NumberFormat nf = NumberFormat.getNumberInstance();
      nf.setMinimumFractionDigits(1);
      nf.setMaximumFractionDigits(1);
      nf.setGroupingUsed(false);
      int cols = 8;
      meanField = new JFormattedTextField(nf);
      meanField.setValue(0d);
      meanField.setColumns(cols);
      meanField.setHorizontalAlignment(JTextField.RIGHT);
      statsPanel.add(meanField);
      corrField = new JFormattedTextField(nf);
      corrField.setValue(0d);
      corrField.setColumns(cols);
      corrField.setHorizontalAlignment(JTextField.RIGHT);
      corrField.setBorder(highlight);
      statsPanel.add(corrField);
      sdField = new JFormattedTextField(nf);
      sdField.setValue(0d);
      sdField.setColumns(cols);
      sdField.setHorizontalAlignment(JTextField.RIGHT);
      statsPanel.add(sdField);
      JButton reMeasureButton = new JButton("Re-Measure");
      reMeasureButton.addActionListener(e -> {
         //We want to measure all channels in the background, without table update
         ImagePlus.removeImageListener(this);
         setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
         SwingWorker<Void, Void> worker = new SwingWorker<Void, Void>() {

            @Override
            protected Void doInBackground() {
               profiler.measureAll();
               return null;
            }

            @Override
            public void done() {
               ImagePlus.addImageListener(Dialog.this);
               profileTableModel.fireTableDataChanged();
               guvTableModel.fireTableDataChanged();
               Dialog.this.setCursor(Cursor.getDefaultCursor());
            }
         };
         worker.run();
      });
      reMeasureButton.setToolTipText("Required if the image has been modified since running the batch " +
            "plugin");
      statsPanel.add(reMeasureButton);
      Dimension d = new Dimension(statsPanel.getPreferredSize().width, statsPanel.getPreferredSize().height);
      statsPanel.setMaximumSize(d);
      cp.add(statsPanel, c);
      constraintReset(c);

      c.gridx = 0;
      c.gridy ++;
      cp.add(Box.createVerticalStrut(6), c);

      //Results
      c.gridy ++;
      c.gridwidth = 2;
      JButton writeButton = new JButton("Save results");
      writeButton.addActionListener(e -> {
         if (profiler.imp != null && profiler.list != null) {
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
            SwingWorker<Void, Void> worker = new SwingWorker<Void, Void>() {

               @Override
               protected Void doInBackground() {
                  ImagePlus imp = profiler.imp;
                  FileInfo info = imp.getOriginalFileInfo();
                  if (info == null) {
                     return null;
                  }
                  Path dir = Paths.get(info.directory);
                  Path zip = dir.resolve(imp.getShortTitle()+".zip");
                  if (Files.exists(zip)) {
                     try {
                        Files.delete(zip);
                     } catch (IOException ex) {
                        SwingUtilities.invokeLater(() -> JOptionPane.showMessageDialog(null,
                              "Could not save ZIP file.",
                              Profiler.TITLE, JOptionPane.ERROR_MESSAGE));
                     }
                  }
                  GUVReader.saveRoiSet(profiler.list, zip.toString());
                  GUVReader.write(profiler.imp, dir, profiler.list, null);
                  return null;
               }

               @Override
               public void done() {
                  Dialog.this.setCursor(Cursor.getDefaultCursor());
               }
            };
            worker.run();
         }
      });
      cp.add(writeButton, c);

      pack();
      setResizable(false);
      Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
      setLocation((screen.width - getWidth())/2, (screen.height - getHeight())/2);
   }

   @Override
   public void imageOpened(ImagePlus imp) {
      if (profiler.imp == null) {
         setImage(imp);
      }
   }

   @Override
   public void imageClosed(ImagePlus imp) {
      if (imp == profiler.imp) {
         imageField.setText("");
         profiler.reset(false);
         guvTableModel.reset();
      }
   }

   @Override
   public void imageUpdated(ImagePlus imp) {
      if (imp != profiler.imp) {
         return;
      }
      profiler.updatePosition(imp.getC(), imp.getZ(), imp.getT());
   }

   @Override
   public void propertyChange(PropertyChangeEvent evt) {
      String name = evt.getPropertyName();
      Object value = evt.getNewValue();
      switch (name) {
         case "mean":
            meanField.setValue(value);
            break;
         case "corrected":
            corrField.setValue(value);
            break;
         case "SD":
            sdField.setValue(value);
            break;
      }
   }

   @Override
   protected void displayAbout() {
      Image icon = Toolkit.getDefaultToolkit().getImage(Dialog.class.getResource("/biooptics.png"));
      JOptionPane.showMessageDialog(this,
            "\u00A9 2021-2022 BioOptics - Light Microscopy, Max Perutz Labs, Vienna\n\n"
                  +"Version "+ version, title, JOptionPane.INFORMATION_MESSAGE, new ImageIcon(icon));
   }

   @Override
   protected void quit() {
      if (profiler.imp != null) {
         if (profiler.imp.getOverlay() != null) {
            profiler.imp.getOverlay().clear();
         }
         profiler.imp.getCanvas().removeKeyListener(keyAdapter);
         profiler.imp.getCanvas().removeMouseListener(mouseAdapter);
         profiler.reset(true);
      }
      ImagePlus.removeImageListener(this);
      GUV.savePreferences();
      super.quit();
   }

   private void add() {
      if (profiler.imp == null) {
         JOptionPane.showMessageDialog(this, "No image set.", title,
               JOptionPane.WARNING_MESSAGE);
         return;
      }
      String tool = IJ.getToolName();
      if (!(tool.equals("Lasso/Blow Tool") || tool.equals("oval"))) {
         JOptionPane.showMessageDialog(this, "Activate drawing mode with 'Draw', " +
               "or draw an oval ROI.", title, JOptionPane.WARNING_MESSAGE);
         return;
      }
      if (profiler.imp.getRoi() == null) {
         JOptionPane.showMessageDialog(this, "No selection.", title,
               JOptionPane.WARNING_MESSAGE);
         return;
      }
      int index = profiler.add(displayOutlines.isSelected());
      if (index > -1) {
         guvTableModel.fireTableDataChanged();
         guvTable.setRowSelectionInterval(index, index);
         guvTable.scrollRectToVisible(guvTable.getCellRect(index, 0, true));
      }
   }

   private void columnWidth(TableColumn column, int columnWidth) {
      column.setMinWidth(columnWidth);
      column.setPreferredWidth(columnWidth);
      column.setMaxWidth(columnWidth);
   }

   private void constraintReset(GridBagConstraints c) {
      c.weightx = 0;
      c.weighty = 0;
      c.gridwidth = 1;
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.LINE_START;
   }

   private void setImage(ImagePlus imp) {
      setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
      if (imp == profiler.imp) {
         return;
      }
      if (profiler.imp != null) {
         imp.getCanvas().removeKeyListener(keyAdapter);
         imp.getCanvas().removeMouseListener(mouseAdapter);
         profiler.reset(false);
      }
      imp.getCanvas().addKeyListener(keyAdapter);
      imp.getCanvas().addMouseListener(mouseAdapter);
      final SwingWorker<Void, Void> worker = new SwingWorker<Void, Void>() {

         @Override
         protected Void doInBackground() {
            profiler.setImage(imp, displayOutlines.isSelected());
            return null;
         }

         @Override
         public void done() {
            imageField.setText(imp.getTitle());
            Dialog.this.setCursor(Cursor.getDefaultCursor());
         }
      };
      worker.run();
   }
}

class HeaderRenderer implements TableCellRenderer {

   private final DefaultTableCellRenderer renderer;

   public HeaderRenderer(JTable table, int horizontalAlignment) {
      renderer = (DefaultTableCellRenderer) table.getTableHeader().getDefaultRenderer();
      renderer.setHorizontalAlignment(horizontalAlignment);
   }

   @Override
   public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
         boolean hasFocus, int row, int col) {
      return renderer.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, col);
   }
}

class DoubleRenderer extends DefaultTableCellRenderer {

   private static final DecimalFormat DF = new DecimalFormat("###.#");

   static {
      DF.setMaximumFractionDigits(1);
      DF.setMinimumFractionDigits(1);
   }

   DoubleRenderer() {
      super();
   }

   @Override
   public void setValue(Object value) {
      setText((value == null) ? "" : DF.format(value));
      setHorizontalAlignment(SwingConstants.RIGHT);
   }
}