package segmentation;

import ij.IJ;
import ij.ImagePlus;
import ij.plugin.FolderOpener;
import ij.plugin.HyperStackConverter;
import ij.plugin.ImagesToStack;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import shared.GUVReader;
import shared.GUVSet;
import shared.IJLogger;

/**Assembles stacks from cellpose output (single-plane) and detects GUVs.*/
public class PostProcessor extends AbstractProcessor {

   PostProcessor(IJLogger logger, BatchInfo info) {
      super(logger, info);
   }

   @Override
   protected Void doInBackground() throws Exception {
      setProgress(0);
      setIncrement(100d/batchInfo.nFiles);
      for (ImageInfo info: batchInfo.infos) {
         if (isCancelled()) {
            break;
         }
         if (!info.exists) {
            ImagePlus masks;
            if (info.isHigherDimensional) {
               /*Note: the following IJ command needs a trailing file separator.
                * The options may change in later IJ versions to " file=_cp_masks.tif"*/
               ImagePlus tmp = FolderOpener.open(info.root.toString() + File.separator,
                     " filter=_cp_masks.tif");
               masks = HyperStackConverter.toHyperStack(tmp, info.nChannels, info.nSlices, info.nFrames,
                     "xytzc", "Color");
               tmp.changes = false;
               tmp.close();
            } else if (info.nChannels > 1) {
               ImagePlus[] imps = new ImagePlus[info.files.size()];
               for (int i = 0; i < info.files.size(); i++) {
                  imps[i] = IJ.openImage(info.files.get(i).toString());
               }
               ImagePlus tmp = ImagesToStack.run(imps);
               masks = HyperStackConverter.toHyperStack(tmp, info.nChannels, info.nSlices, info.nFrames,
                     "xytzc", "Color");
               tmp.changes = false;
               tmp.close();
            } else { //single slice
               masks = IJ.openImage(info.files.get(0).toString());
               masks.setTitle(info.title + "_cp_masks.tif");
            }
            ImagePlus original = IJ.openImage(info.copyOfOriginal.toString());
            ImagePlus resizedMasks = masks.resize(info.originalWidth, info.originalHeight, "none");
            masks.changes = false;
            masks.close();
            resizedMasks.setCalibration(original.getCalibration());

            List<GUVSet> list = GUVReader.readLabels(original, resizedMasks);
            GUVReader.measure(list, original);
            if (batchInfo.dst != null) {
               Path roiSetPath = batchInfo.dst.resolve(info.title+".zip");
               GUVReader.saveRoiSet(list, roiSetPath.toString());
               Path maskPath = batchInfo.dst.resolve(info.title+"_cp_masks.tif");
               if (!IJ.saveAsTiff(resizedMasks, maskPath.toString())) {
                  logger.error("Error while saving "+maskPath);
               }
               GUVReader.write(original, batchInfo.dst, list, logger);
               resizedMasks.changes = false;
               resizedMasks.close();
               original.changes = false;
               original.close();
            } else {
               resizedMasks.setTitle(info.title+"_cp_masks.tif");
               resizedMasks.show();
            }
         }
         incrementProgress();
      }
      GUVReader.deleteDirectory(batchInfo.tmp);
      setProgress(100);
      return null;
   }

   @Override
   public void done() {
      if (Files.exists(batchInfo.tmp)) { //sometimes it is not deleted in the worker thread
         try {
            GUVReader.deleteDirectory(batchInfo.tmp);
         } catch (IOException e) {
            logger.error("Could not delete temporary directory "+batchInfo.tmp);
         }
      }
      String msg;
      if (isCancelled()) {
         msg = "Batch segmentation cancelled";
      } else {
         msg = "Batch segmentation done";
      }
      IJ.showStatus(msg);
   }
}
