package segmentation;

import ij.IJ;
import ij.ImagePlus;
import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import loci.formats.ImageReader;
import loci.formats.MetadataTools;
import loci.formats.meta.IMetadata;
import loci.plugins.BF;
import loci.plugins.in.ImporterOptions;
import ome.xml.model.primitives.Timestamp;
import shared.GUVReader;
import shared.IJLogger;

/**Saves images as single TIFFs in a temporary directory.*/
class PreProcessor extends AbstractProcessor {

   //Bio-Formats format names
   private static final String FORMAT_CZI = "Zeiss CZI";
   private static final String FORMAT_LSM = "Zeiss Laser-Scanning Microscopy";
   private static final String FORMAT_METAMORPH = "Metamorph STK";
   private static final String FORMAT_OME_TIFF = "OME-TIFF";

   private static final Pattern OME_TIFF_STAGE_POSITION =
         Pattern.compile("_\\d+:(.+)_sg:\\d+");
   private static final Pattern METAMORPH_STAGE_POSITION =
         Pattern.compile("Stage\\d+\\s\"(.+)\"");
   private static final Pattern WELL_POSITION =
         Pattern.compile("row:([A-Z]+),column:(\\d+)(?:,site:(\\d+))?");

   private final ImageReader reader;
   private final IMetadata metaData;

   PreProcessor(IJLogger logger, BatchInfo info) {
      super(logger, info);
      reader = new ImageReader();
      metaData = MetadataTools.createOMEXMLMetadata();
      reader.setMetadataStore(metaData);
   }

   @Override
   protected Void doInBackground() throws Exception {
      setProgress(0);
      //Split stacks into single-plane TIFFs and save them in the temporary directory
      if (batchInfo.src == null) { //single open image
         ImagePlus imp = IJ.getImage();
         batchInfo.nFiles = 1;
         copy(imp);
      } else { //batch mode on directory
         File[] files = ImageDirectoryReader.read(batchInfo.src);
         //noinspection ConstantConditions
         batchInfo.nFiles = files.length;
         setIncrement(100d/batchInfo.nFiles);
         for (File file : files) {
            if (isCancelled()) {
               GUVReader.deleteDirectory(batchInfo.tmp);
               return null;
            }
            try {
               //Try to read relevant metadata before opening with the easier BF.openImagePlus() method
               reader.setId(file.getPath());
               String format = reader.getFormat();
               boolean isZeiss = format.equals(FORMAT_CZI) || format.equals(FORMAT_LSM);
               //Transmitted light channel
               int transmittedChannel = 0;
               for (int i = 0; i < metaData.getChannelCount(0); i++) {
                  String channelName = metaData.getChannelName(0, i);
                  if (channelName != null) {
                     if ((isZeiss && channelName.contains("T PMT")) ||
                           (channelName.contains("BF") || channelName.contains("DIC"))) {
                        transmittedChannel = i + 1; //start counting with 1, not zero
                     }
                  }
               }
               //Acquisition date
               String date = "";
               Timestamp timestamp = metaData.getImageAcquisitionDate(0);
               if (timestamp != null) {
                  date = timestamp.toString();
               }
               //Series name (includes positions/wells)
               String shortName = getNameWithoutExtension(file.getName());
               int nSeries = reader.getSeriesCount();
               String[] titles = new String[nSeries];
               if (nSeries == 1) {
                  titles[0] = shortName;
               } else {
                  for (int i = 0; i < nSeries; i++) {
                     reader.setSeries(i);
                     titles[i] = shortName+getSeriesName(metaData.getImageName(i), shortName, format);
                  }
               }
               reader.close();
               ImporterOptions options = new ImporterOptions();
               options.setOpenAllSeries(false);
               options.setId(file.getPath());
               //To open images one by one, all series except a single one have to set OFF (false)
               for (int i = 0; i < nSeries; i++) {
                  options.setSeriesOn(i, false);
               }
               for (int i = 0; i < nSeries; i++) {
                  //Enable just the current series, disable it after opening
                  options.setSeriesOn(i, true);
                  ImagePlus[] imps = BF.openImagePlus(options);
                  if (imps.length != 1) {
                     logger.error("Problem while opening "+file.getName()+". Expected a single image, received "
                           +imps.length);
                  }
                  options.setSeriesOn(i, false);
                  ImagePlus imp = imps[0];
                  if (transmittedChannel > 0) {
                     imp.setProp("transmitted", transmittedChannel);
                  }
                  if (date.isEmpty() && format.equals(FORMAT_CZI)) {
                     //OME-XML does not read the date from Zeiss CZI
                     date = imp.getProp("Information|Document|CreationDate"); //do not use getStringProperty(String)
                  }
                  if (!date.isEmpty()) {
                     imp.setProp("date", date);
                  }
                  imp.setTitle(titles[i]);
                  copy(imp);
                  imp.close();
               }
            } catch (Exception e) {
               logger.error(e);
            }
            incrementProgress();
         }
      }
      setProgress(100);
      return null;
   }

   @Override
   public void done() {
      new CellposeProcessor(logger, batchInfo).execute();
   }

   private static String getNameWithoutExtension(String name) {
      name = name.trim();
      if (name.endsWith(".companion.ome")) {
         return name.substring(0, name.length() - 14);
      } else if (name.endsWith(".ome.tif") || name.endsWith(".ome.btf")) {
         return name.substring(0, name.length() - 8);
      } else {
         return name.substring(0, name.lastIndexOf("."));
      }
   }

   /**Tries to retrieve a series name which is better readable than the default.*
    * @param seriesName               the original series name
    * @param fileNameWithoutExtension the file name without extension
    * @param format                   the format string
    * @return a modified series name (or the original provided)*/
   private static String getSeriesName(String seriesName, String fileNameWithoutExtension, String format) {
      String result;
      switch (format) {
         case FORMAT_METAMORPH:
         case FORMAT_OME_TIFF:
            Matcher m1;
            if (format.equals(FORMAT_METAMORPH)) {
               m1 = METAMORPH_STAGE_POSITION.matcher(seriesName);
            } else {
               seriesName = seriesName.replace(fileNameWithoutExtension, "");
               m1 = OME_TIFF_STAGE_POSITION.matcher(seriesName);
            }
            if (m1.find()) {
               String completeLabel = m1.group(1);
               Matcher m2 = WELL_POSITION.matcher(completeLabel);
               if (m2.find()) {
                  String row = m2.group(1);
                  String column = m2.group(2);
                  String site = m2.group(3);
                  result = "_" + row + column;
                  if (site != null) {
                     result = result + "_P" + site;
                  }
               } else {
                  result = "_" + completeLabel;
               }
            } else {
               result = seriesName;
            }
            break;
         case FORMAT_LSM: {
            String positionName = fileNameWithoutExtension + " #";
            result = seriesName.replace(positionName, "_P");
            break;
         }
         case FORMAT_CZI: {
            //Zeiss ZEN Black und Blue yield different series names (with or without extension)
            String positionName;
            if (seriesName.contains(".czi #")) {
               positionName = fileNameWithoutExtension + ".czi #";
            } else {
               positionName = fileNameWithoutExtension + " #";
            }
            result = seriesName.replace(positionName, "_P");
            break;
         }
         default:
            result = seriesName;
            break;
      }
      return result;
   }

   private void copy(ImagePlus imp) throws IOException {
      batchInfo.nPlanes += imp.getStackSize();
      ImageInfo imageInfo = new ImageInfo(batchInfo, imp);
      if (!imageInfo.isMicrons) {
         logger.warning(String.format("Image %s is not scaled in microns. Will be skipped",
               imageInfo.title));
         return;
      }
      ImagePlus black = IJ.createImage("T-PMT", imageInfo.type+" black", imageInfo.rescaledWidth,
            imageInfo.rescaledHeight, 1);
      int transmittedInfo = batchInfo.transmittedChannel; //0 = auto-detect, -1 = none, or the channel
      int autodetected = -1;
      String str = imp.getProp("transmitted"); //do not use getStringProperty(String) - does not work
      if ((str != null)) {
         try {
            autodetected = Integer.parseInt(str);
         } catch (NumberFormatException ex) {
            //ignored
         }
      }
      for (int c = 1; c <= imp.getNChannels(); c++) {
         boolean isTransmitted = false;
         if (transmittedInfo == 0) { //autodetect the transmitted light channel
            if (c == autodetected) {
               isTransmitted = true;
            }
         } else if (transmittedInfo > 0) { //transmitted light channel specified by user
            if (c == transmittedInfo) {
               isTransmitted = true;
               imp.setProp("transmitted", c);  //do not use setProperty(String, Object) - does not work
            }
         }
         for (int z = 1; z <= imp.getNSlices(); z++) {
            for (int t = 1; t <= imp.getNFrames(); t++) {
               String title = imp.getShortTitle();
               if (imp.getNChannels() > 1) {
                  title = title+"_C"+c;
               }
               if (imp.getNSlices() > 1) {
                  title = title+"_Z"+z;
               }
               if (imp.getNFrames() > 1) {
                  title = title+"_T"+t;
               }
               if (isTransmitted) {
                  saveSingleSlice(black, imageInfo, title, false);
               } else {
                  ImagePlus imp2 = GUVReader.duplicate(imp, c, c, z, z, t, t);
                  saveSingleSlice(imp2, imageInfo, title, true);
                  imp2.changes = false;
                  imp2.close();
               }
            }
         }
      }
      black.close();
      if (batchInfo.dst != null) {
         IJ.saveAsTiff(imp, imageInfo.copyOfOriginal.toString());
      }
   }

   /**Saves a single slice
    * @param title short title (without extension)*/
   private void saveSingleSlice(ImagePlus imp, ImageInfo info, String title, boolean resize) {
      Path path = info.root.resolve(title + ".tif");
      if (resize) {
         ImagePlus imp2 = imp.resize(info.rescaledWidth, info.rescaledHeight, "none");
         IJ.run(imp2, "16-bit", "");
         IJ.run(imp2, "Enhance Contrast...", "saturated=1");
         IJ.saveAsTiff(imp2, path.toString());
         imp2.changes = false;
         imp2.close();
      } else {
         IJ.saveAsTiff(imp, path.toString());
      }
      if (!info.isHigherDimensional) {
         info.files.add(info.root.resolve(title + "_cp_masks.tif"));
      }
   }
}
