package shared;

import ch.epfl.biop.MaxInscribedCircles;
import fiji.util.Circle_Fitter;
import ij.IJ;
import ij.ImagePlus;
import ij.gui.Line;
import ij.gui.OvalRoi;
import ij.gui.Roi;
import ij.measure.Calibration;
import ij.measure.ResultsTable;
import ij.plugin.Duplicator;
import ij.plugin.RoiEnlarger;
import ij.plugin.frame.RoiManager;
import ij.process.ImageProcessor;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.stream.IntStream;
import javax.swing.JOptionPane;
import profiler.Profiler;

/**Static methods to handle GUVs.*/
public final class GUVReader {

   private static final Circle_Fitter fitter = new Circle_Fitter();
   private static final Duplicator duplicator = new Duplicator();
   //private static final RoiManager rm = RoiManager.getRoiManager(); //visible manager for testing
   private static final RoiManager rm = new RoiManager(false); //ctor version, hidden

   private static final DateTimeFormatter DTF = DateTimeFormatter.ofPattern("yyyy-MM-dd");

   private GUVReader() {}

   /**Recursively deletes the content of a directory.
    * @param path the path of the directory*/
   public static void deleteDirectory(Path path) throws IOException {
      if (Files.isDirectory(path, LinkOption.NOFOLLOW_LINKS)) {
         try (DirectoryStream<Path> entries = Files.newDirectoryStream(path)) {
            for (Path entry : entries) {
               deleteDirectory(entry);
            }
         }
      }
      Files.delete(path);
   }

   /**Duplicates an image.
    * @param imp the image
    * @param firstC the first channel
    * @param lastC the last channel
    * @param firstZ the first z-scice
    * @param lastZ the last z-slice
    * @param firstT the first frame (time point)
    * @param lastT the last frame (time point)*/
   public static ImagePlus duplicate(ImagePlus imp, int firstC, int lastC, int firstZ, int lastZ,
         int firstT, int lastT) {
      return duplicator.run(imp, firstC, lastC, firstZ, lastZ, firstT, lastT);
   }

   /**Fits a circle to a region of interest in an image.
    * @param imp the image
    * @param roi the region
    * @param pixelsPerMicron rounded to an integer
    * @param show displays the temporary window created for fitting*/
   public static OvalRoi fitCircle(ImagePlus imp, Roi roi, int pixelsPerMicron, boolean show) {
      /*Measures the average intensity in a central region and compares it to the entire ROI.
        More or less homogenous beads (without sharp border) should be enlarged for detection/fitting.*/
      Rectangle r = roi.getBounds();
      float centerW = r.width;
      float centerH = r.height;
      Roi center = new Roi(r.getCenterX(), r.getCenterY(), centerW/4, centerH/4);
      imp.setRoi(center);
      double meanCenter = center.getStatistics().mean;
      imp.setRoi(roi);
      double meanAll = roi.getStatistics().mean;
      double ratio = meanAll/meanCenter;
      if (ratio < 1.2d) {
         RoiEnlarger.enlarge(imp, pixelsPerMicron * 3);
         roi = imp.getRoi();
      }
      double originalX = roi.getXBase();
      double originalY = roi.getYBase();

      //Create a temporary image, copy the ROI into it and fit a circle
      int width = 512;
      if (Math.max(roi.getBounds().width, roi.getBounds().height) > 256) {
         width = 1024;
      }
      @SuppressWarnings("SuspiciousNameCombination")
      ImagePlus tmp = IJ.createImage("TEMP", "16-bit black", width, width, 1);
      if (show) {
         tmp.show();
      }
      imp.setRoi(roi, true);
      imp.copy();
      tmp.paste();
      imp.deleteRoi();
      double tempX = tmp.getRoi().getXBase(); //upper left corner X of the enlarged ROI in the temp image
      double tempY = tmp.getRoi().getYBase(); //upper left corner Y
      tmp.deleteRoi();
      IJ.run(tmp, "Enhance Contrast...", "saturated=0.1");
      fitter.setImageProcessor(tmp.getProcessor());
      fitter.setAutoThreshold();
      OvalRoi oval = fitter.calculateRoi();
      if (show) {
         tmp.setRoi(oval);
      }
      //Get the offset of the origin of the circle relative to the ROI pasted in the temporary image
      double offsetX = oval.getXBase() - tempX;
      double offsetY = oval.getYBase() - tempY;
      //Shift the location of the circle to the one in the original image, corrected by the offset
      oval.setLocation(originalX + offsetX, originalY + offsetY);
      //Cleanup
      tmp.changes = false;
      tmp.close();
      return oval;
   }

   /**Returns the length of the provided distance (in µm) in pixels.
    * @param cal the calibration of the image
    * @param microns the distance in microns
    * @param defaultValue returned if the image is not calibrated in microns
    * @return the pixel number, or the default value*/
   public static double getPixelsForMicrons(Calibration cal, double microns, double defaultValue) {
      if (cal.getUnit().equals("micron") || cal.getUnit().equals("µm") || cal.getUnit().equals("um")) {
         return microns/cal.pixelWidth;
      } else {
         return defaultValue;
      }
   }

   /**Returns the area (in pixels) for a circle with the given diameter (in µm).
    * @param cal the calibration of the image
    * @param diameter the area in square microns
    * @param defaultValue returned if the image is not calibrated in microns
    * @return the pixel number, or the default value*/
   public static int getPixelsForDiameter(Calibration cal, double diameter, int defaultValue) {
      if (cal.getUnit().equals("micron") || cal.getUnit().equals("µm") || cal.getUnit().equals("um")) {
         double disk = Math.pow(diameter/2, 2) * Math.PI;
         double pixelArea = cal.pixelWidth * cal.pixelHeight;
         return (int) (disk/pixelArea);
      } else {
         return defaultValue;
      }
   }

   /**Compares the location of 2 oval ROIs.
    * @param a the first ROI
    * @param b the second ROI
    * @param intersecting fraction of minimum overlap (of the smaller ROI; e.g. 0.8)
    * @return the larger ROI if they overlap, or <code>null</code> if the do not*/
   public static Roi isOverlapping(Roi a, Roi b, double intersecting) {
      Roi larger = a;
      Roi smaller = b;
      if (b.getStatistics().area > a.getStatistics().area) {
         larger = b;
         smaller = a;
      }
      int intersectingPixels = 0;
      Point[] points = larger.getContainedPoints();
      for (Point p : points) {
         if (smaller.contains(p.x, p.y)) {
            intersectingPixels ++;
         }
      }
      if (intersectingPixels == 0) {
         return null;
      }
      double intersectingFraction = (double) intersectingPixels/ (double) smaller.getContainedPoints().length;
      if (intersectingFraction > intersecting) {
         return larger;
      } else {
         return null;
      }
   }

   /**Measures all GUVs in all positions (frames, slices, channels) of the image provided.
    * @param list a list of GUV sets
    * @param imp the image*/
   public static void measure(List<GUVSet> list, ImagePlus imp) {
      for (int i = 1; i <= imp.getNChannels() ; i++) {
         for (GUVSet set : list) {
            imp.setPosition(i, set.slice, set.frame);
            set.measure(imp);
         }
      }
   }

   /**Reads labels (masks) from a label image.
    * @param imp the image
    * @param masks the mask image*/
   public static List<GUVSet> readLabels(ImagePlus imp, ImagePlus masks) {
      List<GUVSet> list = new ArrayList<>(imp.getNFrames() * imp.getNSlices());
      for (int i = 1; i <= masks.getNFrames(); i++) {
         for (int j = 1; j <= masks.getNSlices() ; j++) {
            //Channel-independent GUV collection (all circles from all channels) for each slice and frame
            rm.reset();
            List<OvalRoi> circles = new ArrayList<>();
            for (int k = 1; k <= masks.getNChannels(); k++) {
               imp.setPosition(k, j, i);
               masks.setPosition(k, j, i);
               ImageProcessor ip = masks.getProcessor();
               ip.setThreshold(1, ip.getMax(), ImageProcessor.NO_LUT_UPDATE);
               IJ.run(masks, "Create Selection", "");
               Roi roi = masks.getRoi();
               if (roi != null) {
                  rm.addRoi(masks.getRoi());
                  masks.deleteRoi();
                  threshold(imp, masks, circles, k != 1);
               }
            }
            Roi all;
            if (rm.getCount() == 0) {
               continue;
            } else if (rm.getCount() == 1) {
               all = rm.getRoi(0);
            } else {
               rm.setSelectedIndexes(IntStream.rangeClosed(0, rm.getCount() - 1).toArray());
               rm.runCommand(masks,"Combine");
               all = masks.getRoi();
               masks.deleteRoi();
            }
            all.setName("All-Z"+j+"T"+i);
            all.setPosition(0, j, i);
            GUVSet set = new GUVSet(i, j, imp.getNChannels(), all);
            list.add(set);
            rm.reset();
            for (int k = 0; k < circles.size(); k++) {
               set.add(new GUV(k+1, imp, circles.get(k), j, i, all));
            }
            //Construct a control/noise GUV in an empty space
            imp.setRoi(all);
            IJ.run(imp, "Make Inverse", "");
            List<Roi> inscribedList = MaxInscribedCircles.findCircles(imp, 0, true);
            if (inscribedList.size() == 1) {
               Roi inscribed = inscribedList.get(0);
               double d0 = inscribed.getBounds().width * imp.getCalibration().pixelWidth;
               double shrinkMicrons = 2 * GUV.profileExt + 5; //make it smaller to avoid neighbouring GUVs
               double d1 = d0 - shrinkMicrons; //new diameter after shrinking
               if (d1 > 180) { //don't make it unusually large
                  double delta = d1 - 180;
                  shrinkMicrons += delta;
               }
               if (d0 > 10) { //don't set a very small background circle
                  int shrinkPixels = (int) - getPixelsForMicrons(imp.getCalibration(), shrinkMicrons,25);
                  imp.setRoi(inscribed);
                  RoiEnlarger.enlarge(imp, shrinkPixels);
                  OvalRoi resized = (OvalRoi) imp.getRoi();
                  GUV control = new GUV(circles.size() + 1, imp, resized, j, i, all);
                  control.setAsControl(true);
                  set.add(control);
               }
            }
         }
      }
      return list;
   }

   /**Reads a zipped ROI set.
    * @param imp the image
    * @param path the path of the ROI set*/
   public static List<GUVSet> readRoiSet(ImagePlus imp, String path) {
      List<GUVSet> list = new ArrayList<>(imp.getNFrames() * imp.getNSlices());
      for (int i = 1; i <= imp.getNFrames(); i++) {
         for (int j = 1; j <= imp.getNSlices() ; j++) {
            list.add(new GUVSet(i, j, imp.getNChannels()));
         }
      }
      rm.reset();
      rm.runCommand("Open", path);
      for (int i = 0; i < rm.getCount(); i++) {
         Roi roi = rm.getRoi(i);
         GUVSet set = getGUVSet(list, roi.getTPosition(), roi.getZPosition());
         if (set == null) {
            continue;
         }
         if (roi instanceof OvalRoi) {
            set.add(new GUV((OvalRoi) roi));
         } else if (roi instanceof Line) {
            //noinspection UnnecessaryContinue
            continue; //lines/profiles have to be processed in a second round (once all GUVs exist)
         } else {
            set.all = roi;
         }
      }
      for (int i = 0; i < rm.getCount(); i++) {
         Roi roi = rm.getRoi(i);
         if (roi instanceof Line) {
            Line line = (Line) roi;
            int n = Integer.parseInt(line.getProperty("GUV-id")) - 1;
            GUVSet set = getGUVSet(list, line.getTPosition(), line.getZPosition());
            if (set == null) {
               continue;
            }
            GUV guv = set.get(n);
            if (guv != null) {
               guv.addProfile(line);
            }
         }
      }
      for (GUVSet set : list) {
         set.calculateAll();
      }
      return list;
   }

   /**From a list of GUV sets, this method retrieves the set of the frame and slice position provided.*/
   private static GUVSet getGUVSet(List<GUVSet> list, int frame, int slice) {
      for (GUVSet set : list) {
         if (set.frame == frame && set.slice == slice) {
            return set;
         }
      }
      return null;
   }

   /**Saves a list of GUV sets to a ZIP file.
    * @param list the list
    * @param path the file path
    * @return <code>true</code> if there are ROIs to be saved and saving succeeds, <code>false</code>
    * otherwise*/
   @SuppressWarnings("UnusedReturnValue")
   public static boolean saveRoiSet(List<GUVSet> list, String path) {
      rm.reset();
      if (list.size() == 0) {
         return false;
      }
      for (GUVSet set : list) {
         if (set.all != null) {
            rm.addRoi(set.all);
         }
         for (GUV guv : set.getAll()) {
            rm.addRoi(guv.roi);
            for (Profile p : guv.profiles) {
               rm.addRoi(p.line);
            }
         }
      }
      return rm.runCommand("Save", path);
   }

   /**Writes the results of a {@link GUVSet} to disk.
    * @param imp the image
    * @param dst the destination directory
    * @param list the {@link GUVSet} list associated with the image
    * @param logger the logger. If this is <code>null</code>, error messages are displayed to users.*/
   public static void write(ImagePlus imp, Path dst, List<GUVSet> list, IJLogger logger) {
      int transmittedChannel = -1;
      Object transmittedObj = imp.getProp("transmitted"); //do not use imp.getProperty(String)
      if (transmittedObj != null) {
         transmittedChannel = Integer.parseInt(transmittedObj.toString()) - 1;
      }
      String date = null;
      Object dateObj = imp.getProp("date");
      if (dateObj != null) {
         LocalDateTime dateTime = LocalDateTime.parse(dateObj.toString());
         date = DTF.format(dateTime);
      }
      for (GUVSet set : list) {
         StringBuilder sb = new StringBuilder(imp.getShortTitle());
         if (imp.getNSlices() > 1) {
            sb.append("_Z");
            sb.append(set.slice + 1);
         }
         if (imp.getNFrames() > 1) {
            sb.append("_T");
            sb.append(set.frame + 1);
         }
         sb.append("_C");
         for (int i = 0; i < set.mean.length; i++) {
            if (i == transmittedChannel) {
               continue;
            }
            String title = sb+String.valueOf(i + 1);
            if (date != null) {
               title = title+"_"+date;
            }
            title += ".csv";
            Path path = dst.resolve(title);
            ResultsTable table = new ResultsTable();
            if (imp.getBitDepth() == 8) {
               table.setPrecision(2);
            } else {
               table.setPrecision(1);
            }
            table.setPrecision(1);
            double control = 0d;
            if (set.getControl() != null) {
               control = set.getControl().mean[i];
            }
            for (GUV guv : set.getAll()) {
               if (guv.isUsed() && !guv.isControl()) {
                  table.addRow();
                  table.addValue("#", guv.id);
                  table.addValue("Profiles", guv.getNumberOfUsedProfiles());
                  double value = guv.mean[i];
                  table.addValue("Raw", value);
                  table.addValue("Corrected", value - control);
                  table.addValue("SD", guv.sd[i]);
               }
            }
            boolean result = table.save(String.valueOf(path));
            if (!result) {
               String msg = "Could not save\n"+path+".";
               if (logger == null) {
                  JOptionPane.showMessageDialog(null, msg, Profiler.TITLE,
                        JOptionPane.ERROR_MESSAGE);
               } else {
                  logger.error(msg);
               }
            }
         }
      }
   }

   /**Thresholds ROIs from a label (mask) image at the given stack position, fits circles and adds them to a
    * list of ROIs.
    * @param imp the original image
    * @param masks the label image
    * @param list the list
    * @param compare indicates to check for similar positions, returning a list with unique ROIs*/
   private static void threshold(ImagePlus imp, ImagePlus masks, List<OvalRoi> list, boolean compare) {
      int pixelPerMicron = (int) Math.round(getPixelsForMicrons(imp.getCalibration(), 1, 1));
      ImageProcessor ip = masks.getProcessor();
      int[] histogram = ip.getHistogram();
      for (int i = 1; i <= ip.getMax(); i++) {
         int value = histogram[i];
         if (value > getPixelsForDiameter(imp.getCalibration(), 12, 2500)) {
            ip.setThreshold(i, i, ImageProcessor.NO_LUT_UPDATE);
            IJ.run(masks, "Create Selection", "");
            OvalRoi newCircle = fitCircle(imp, masks.getRoi(), pixelPerMicron, false);
            masks.deleteRoi();
            //Set slice and frame position (but no channel position)
            newCircle.setPosition(0, masks.getZ(), masks.getT());
            if (compare) {
               boolean isNew = true;
               ListIterator<OvalRoi> it = list.listIterator();
               while (it.hasNext()) {
                  OvalRoi existing = it.next();
                  OvalRoi larger = (OvalRoi) isOverlapping(newCircle, existing, 0.8d);
                  if (larger != null) {
                     it.set(larger);
                     isNew = false;
                     break;
                  }
               }
               if (isNew) {
                  list.add(newCircle);
               }
            } else {
               list.add(newCircle);
            }
         }
      }
   }
}
